#include "board.h"

extern phrase cardback;
extern phrase cursorgfx;

extern phrase atari800;
extern phrase atari2600;
extern phrase atari5200;
extern phrase atari7800;
extern phrase xegs;
extern phrase lynx;
extern phrase jaguar;
extern phrase atarist;

extern phrase atari800game;
extern phrase atari2600game;
extern phrase atari5200game;
extern phrase atari7800game;
extern phrase xegsgame;
extern phrase lynxgame;
extern phrase jaguargame;
extern phrase ataristgame;

typedef enum card {
	ATARI2600 = 0,
	ATARI5200 = 1,
	ATARI7800 = 2,
	LYNX = 3,
	JAGUAR = 4,
	ATARI800 = 5,
	ATARIST = 6,
	XEGS = 7,
	
	ATARI2600_GAME = 8,
	ATARI5200_GAME = 9,
	ATARI7800_GAME = 10,
	LYNX_GAME = 11,
	JAGUAR_GAME = 12,
	ATARI800_GAME = 13,
	ATARIST_GAME = 14,
	XEGS_GAME = 15
} card;

typedef struct cell {
	sprite* faceUpSprite;
	sprite* faceDownSprite;
	card value;
	int isFaceUp;
	int isHidden;
} cell;

typedef struct cursor {
	sprite* spr;
	int row, col;
} cursor;

#define CELL_SIZE_PIXELS 48
#define NUM_CELL_ROWS 4
#define NUM_CELL_COLS 4
#define NUM_CELLS NUM_CELL_ROWS * NUM_CELL_COLS

cell p_cells[NUM_CELL_ROWS][NUM_CELL_COLS];
cursor p_cursor;
int p_movesCount;

// element at 0 holds the penultimately revealed cell
// element at 1 holds the lastly revealed cell
cell* p_lastCells[2];

cell* p_cellBeingMadeFaceDown;
cell* p_cellBeingMadeFaceUp;

int p_cellsMatch( cell* first, cell* second ) {
	if( !first->isFaceUp || !second->isFaceUp ) {
		return 0;
	}
	
	int diff1 = first->value - second->value;
	int diff2 = second->value - first->value;
	
	// since card value (k) matches card value (k + NUM_CELLS/2)
	// see card enum definition for more clarification
	return (diff1 == NUM_CELLS/2) || (diff2 == NUM_CELLS/2);
}

void p_getCellCoordsByIndex( int row, int col, int* x, int* y ) {
	#define INTER_CELL_PADDING 8
	#define LEFT_PADDING 40
	#define TOP_PADDING 33
	*x = LEFT_PADDING + col * ( CELL_SIZE_PIXELS + INTER_CELL_PADDING );
	*y = TOP_PADDING + row * ( CELL_SIZE_PIXELS + INTER_CELL_PADDING );
}

phrase* p_getTextureByCard( card c ) {
	switch( c ) {
		case ATARI2600:
			return &atari2600;
		case ATARI5200:
			return &atari5200;
		case ATARI7800:
			return &atari7800;
		case LYNX:
			return &lynx;
		case JAGUAR:
			return &jaguar;
		case ATARI800:
			return &atari800;
		case ATARIST:
			return &atarist;
		case XEGS:
			return &xegs;
		case ATARI2600_GAME:
			return &atari2600game;
		case ATARI5200_GAME:
			return &atari5200game;
		case ATARI7800_GAME:
			return &atari7800game;
		case LYNX_GAME:
			return &lynxgame;
		case JAGUAR_GAME:
			return &jaguargame;
		case ATARI800_GAME:
			return &atari800game;
		case ATARIST_GAME:
			return &ataristgame;
		case XEGS_GAME:
			return &xegsgame;
	}
	
	return NULL;
}

sprite* p_getCurrentlyVisibleSprite( cell* c ) {
	return c->isFaceUp ? c->faceUpSprite : c->faceDownSprite;
}

sprite* p_getCurrentlyInvisibleSprite( cell* c ) {
	return c->isFaceUp ? c->faceDownSprite : c->faceUpSprite;
}

void p_showFaceUp( cell* c ) {
	c->isFaceUp = 1;
	c->faceUpSprite->invisible = 0;
	c->faceDownSprite->invisible = 1;
}

void p_showFaceDown( cell* c ) {
	c->isFaceUp = 0;
	c->faceUpSprite->invisible = 1;
	c->faceDownSprite->invisible = 0;
}

void p_hide( cell* c ) {
	c->isHidden = 1;
	c->faceUpSprite->invisible = 1;
	c->faceDownSprite->invisible = 1;
}

void p_refreshCursor() {
	int x, y;
	
	p_getCellCoordsByIndex( p_cursor.row, p_cursor.col, &x, &y );
	p_cursor.spr->x = x;
	p_cursor.spr->y = y;
}

cell* p_getHighlightedCell() {
	return &p_cells[p_cursor.row][p_cursor.col];
}

void board_init() {
	int row, col;
	int x, y, i;
	
	int shuffledCardValues[NUM_CELLS];
	
	for( i = 0; i < NUM_CELLS; i++ ) {
		shuffledCardValues[i] = i;
	}
	shuffleInPlace( shuffledCardValues, NUM_CELLS );
	
	for( row = 0; row < NUM_CELL_ROWS; row++ ) {
		for( col = 0; col < NUM_CELL_COLS; col++ ) {
			p_cells[row][col].isHidden = 0;
			p_cells[row][col].value = shuffledCardValues[ row * NUM_CELL_COLS + col ];
			
			p_getCellCoordsByIndex( row, col, &x, &y );
			
			// each of the cells needs a face down sprite
			p_cells[row][col].faceDownSprite = game_makeSprite( CELL_SIZE_PIXELS, CELL_SIZE_PIXELS, &cardback );
			p_cells[row][col].faceDownSprite->x = x;
			p_cells[row][col].faceDownSprite->y = y;
			
			p_cells[row][col].faceUpSprite = game_makeSprite( 
				CELL_SIZE_PIXELS, 
				CELL_SIZE_PIXELS, 
				p_getTextureByCard( p_cells[row][col].value )
				);
			p_cells[row][col].faceUpSprite->x = x;
			p_cells[row][col].faceUpSprite->y = y;
			
			p_showFaceDown( &p_cells[row][col] );
		}
	}
	
	p_cursor.spr = game_makeSprite( 56, 56, &cursorgfx );
	p_cursor.row = 0;
	p_cursor.col = 0;
	p_refreshCursor();
	
	animation_init();
	
	p_lastCells[0] = NULL;
	p_lastCells[1] = NULL;
	p_cellBeingMadeFaceDown = NULL;
	p_cellBeingMadeFaceUp = NULL;
	
	p_movesCount = 0;
}

int p_isVictory() {
	int row, col;
	int numHidden = 0;
	
	for( row = 0; row < NUM_CELL_ROWS; row++ ) {
		for( col = 0; col < NUM_CELL_COLS; col++ ) {
			if( p_cells[row][col].isHidden ) {
				numHidden++;
			}
		}
	}
	
	return numHidden == NUM_CELLS;
}

void board_cursorAction( cursoraction action ) {
	
	cell* current = p_getHighlightedCell();
	
	if( p_isVictory() ) {
		return;
	}
	
	if( animation_isRunning() ) {
		return;
	}
	
	switch( action ) {
		case UP:
			p_cursor.row--;
			break;
		case DOWN:
			p_cursor.row++;
			break;
		case LEFT:
			p_cursor.col--;
			break;
		case RIGHT:
			p_cursor.col++;
			break;
		case FIRE:
			if( current->isHidden || current->isFaceUp ) {
				break;
			}
			p_cellBeingMadeFaceUp = p_getHighlightedCell();
			
			if( p_lastCells[0] != NULL ) {
				p_cellBeingMadeFaceDown = p_lastCells[0];
				animation_start( 
					p_getCurrentlyVisibleSprite( p_cellBeingMadeFaceDown ),
					p_getCurrentlyInvisibleSprite( p_cellBeingMadeFaceDown )
					);
			} else {
				animation_start( 
					p_getCurrentlyVisibleSprite( p_cellBeingMadeFaceUp ),
					p_getCurrentlyInvisibleSprite( p_cellBeingMadeFaceUp )
					);
			}
			
			p_lastCells[0] = p_lastCells[1];
			p_lastCells[1] = p_cellBeingMadeFaceUp;
			
			p_movesCount++;
			break;
	}
	
	if( p_cursor.row < 0 ) {
		p_cursor.row = NUM_CELL_ROWS - 1;
	}
	if( p_cursor.row >= NUM_CELL_ROWS ) {
		p_cursor.row = 0;
	}
	
	if( p_cursor.col < 0 ) {
		p_cursor.col = NUM_CELL_COLS - 1;
	}
	if( p_cursor.col >= NUM_CELL_COLS ) {
		p_cursor.col = 0;
	}
	
	p_refreshCursor();
}

void board_frame() {
	
	cell* temp;
	
	if( p_isVictory() ) {
		p_cursor.spr->invisible = 1;
	}
	
	animation_advance();
	if( !animation_hasJustFinished() ) {
		return;
	}
	
	// were we animating a cell going face down?
	if( p_cellBeingMadeFaceDown != NULL ) {
		p_cellBeingMadeFaceDown->isFaceUp = 0;
		p_cellBeingMadeFaceDown = NULL;
		// we finished animating a face down
		// now animate face up on the currently highlighted cell
		animation_start( 
			p_getCurrentlyVisibleSprite( p_cellBeingMadeFaceUp ),
			p_getCurrentlyInvisibleSprite( p_cellBeingMadeFaceUp )
			);
		return;
	} 
	
	// we finished animating a cell going face up
	p_cellBeingMadeFaceUp->isFaceUp = 1;
	temp = p_cellBeingMadeFaceUp;
	p_cellBeingMadeFaceUp = NULL;

	// check for a match
	if( p_lastCells[0] != NULL && p_lastCells[1] != NULL &&
		p_cellsMatch( p_lastCells[0], p_lastCells[1] )
		) {
		
		p_hide( p_lastCells[0] );
		p_hide( p_lastCells[1] );
		
		p_lastCells[0] = NULL;
		p_lastCells[1] = NULL;
	}
}

int board_getMovesCount() {
	return p_movesCount;
}

